import enum
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    NamedTuple,
    Callable,
    Generator,
    Iterator,
    Tuple,
    TYPE_CHECKING,
    TypeVar,
)
from opentrons.protocol_api.labware import Well
from opentrons import types
from opentrons.protocols.api_support.types import APIVersion

if TYPE_CHECKING:
    from opentrons.protocol_api.contexts import InstrumentContext
    from opentrons.protocols.execution.dev_types import Dictable


class MixStrategy(enum.Enum):
    BOTH = enum.auto()
    BEFORE = enum.auto()
    AFTER = enum.auto()
    NEVER = enum.auto()


class DropTipStrategy(enum.Enum):
    TRASH = enum.auto()
    RETURN = enum.auto()


class TouchTipStrategy(enum.Enum):
    NEVER = enum.auto()
    ALWAYS = enum.auto()


class BlowOutStrategy(enum.Enum):
    NONE = enum.auto()
    TRASH = enum.auto()
    DEST = enum.auto()
    SOURCE = enum.auto()
    CUSTOM_LOCATION = enum.auto()


class TransferMode(enum.Enum):
    DISTRIBUTE = enum.auto()
    CONSOLIDATE = enum.auto()
    TRANSFER = enum.auto()


class Transfer(NamedTuple):
    """
    Options pertaining to behavior of the transfer.

    """

    new_tip: types.TransferTipPolicy = types.TransferTipPolicy.ONCE
    air_gap: float = 0
    carryover: bool = True
    gradient_function: Optional[Callable[[float], float]] = None
    disposal_volume: float = 0
    mix_strategy: MixStrategy = MixStrategy.NEVER
    drop_tip_strategy: DropTipStrategy = DropTipStrategy.TRASH
    blow_out_strategy: BlowOutStrategy = BlowOutStrategy.NONE
    touch_tip_strategy: TouchTipStrategy = TouchTipStrategy.NEVER


Transfer.new_tip.__doc__ = """
    Control when or if to pick up tip during a transfer

    :py:attr:`types.TransferTipPolicy.ALWAYS`
        Drop and pick up a new tip after each dispense.

    :py:attr:`types.TransferTipPolicy.ONCE`
        Pick up tip at the beginning of the transfer and use it
        throughout the transfer. This would speed up the transfer.

    :py:attr:`types.TransferTipPolicy.NEVER`
        Do not ever pick up or drop tip. The protocol should explicitly
        pick up a tip before transfer and drop it afterwards.

    To customize where to drop tip, see :py:attr:`.drop_tip_strategy`.
    To customize the behavior of pickup tip, see
    :py:attr:`.TransferOptions.pick_up_tip`.
    """

Transfer.air_gap.__doc__ = """
    Controls the volume (in uL) of air gap aspirated when moving to
    dispense.

    Adding an air gap would slow down a transfer since less liquid will
    now fit in the pipette but it prevents the loss of liquid while
    moving between wells.
    """

Transfer.carryover.__doc__ = """
    Controls whether volumes larger than pipette's max volume will be
    split into smaller volumes.
    """

Transfer.gradient_function.__doc__ = """
    Specify a nonlinear gradient for volumes.

    This should be a function that takes a single float between 0 and 1
    and returns a single float between 0 and 1. This function is used
    to determine the path the transfer takes between the volume
    gradient minimum and maximum if the transfer volume is specified as
    a gradient. For instance, specifying the function as

    .. code-block:: python

        def gradient(a):
            if a > 0.5:
                return 1.0
            else:
                return 0.0

    would transfer the minimum volume of the gradient to the first half
    of the target wells, and the maximum to the other half.
    """

Transfer.disposal_volume.__doc__ = """
    The amount of liquid (in uL) to aspirate as a buffer.

    The remaining buffer will be blown out into the location specified
    by :py:attr:`.blow_out_strategy`.

    This is useful to avoid under-pipetting but can waste reagent and
    slow down transfer.
    """

Transfer.mix_strategy.__doc__ = """
    If and when to mix during a transfer.

    :py:attr:`MixStrategy.NEVER`
        Do not ever perform a mix during the transfer.

    :py:attr:`MixStrategy.BEFORE`
        Mix before each aspirate.

    :py:attr:`MixStrategy.AFTER`
        Mix after each dispense.

    :py:attr:`MixStrategy.BOTH`
        Mix before each aspirate and after each dispense.

    To customize the mix behavior, see :py:attr:`.TransferOptions.mix`
    """

Transfer.drop_tip_strategy.__doc__ = """
    Specifies the location to drop tip into.

    :py:attr:`DropTipStrategy.TRASH`
        Drop the tip into the trash container.

    :py:attr:`DropTipStrategy.RETURN`
        Return the tip to tiprack.
    """

Transfer.blow_out_strategy.__doc__ = """
    Specifies the location to blow out the liquid in the pipette to.

    :py:attr:`BlowOutStrategy.TRASH`
        Blow out to trash container.

    :py:attr:`BlowOutStrategy.SOURCE`
        Blow out into the source well in order to dispense any leftover
        liquid.

    :py:attr:`BlowOutStrategy.DEST`
        Blow out into the destination well in order to dispense any leftover
        liquid.

    :py:attr:`BlowOutStrategy.CUSTOM_LOCATION`
        If using any other location to blow out to. Specify the location in
        :py:attr:`.TransferOptions.blow_out`.
    """

Transfer.touch_tip_strategy.__doc__ = """
    Controls whether to touch tip during the transfer

    This helps in getting rid of any droplets clinging to the pipette
    tip at the cost of slowing down the transfer.

    :py:attr:`TouchTipStrategy.NEVER`
        Do not touch tip ever during the transfer.

    :py:attr:`TouchTipStrategy.ALWAYS`
        Touch tip after each aspirate.

    To customize the behavior of touch tips, see
    :py:attr:`.TransferOptions.touch_tip`.
    """


class PickUpTipOpts(NamedTuple):
    """
    Options to customize :py:attr:`.Transfer.new_tip`.

    These options will be passed to
    :py:meth:`InstrumentContext.pick_up_tip` when it is called during
    the transfer.
    """

    location: Optional[types.Location] = None
    presses: Optional[int] = None
    increment: Optional[int] = None


PickUpTipOpts.location.__doc__ = ":py:class:`types.Location`"
PickUpTipOpts.presses.__doc__ = ":py:class:`int`"
PickUpTipOpts.increment.__doc__ = ":py:class:`int`"


class MixOpts(NamedTuple):
    """
    Options to customize behavior of mix.

    These options will be passed to
    :py:meth:`InstrumentContext.mix` when it is called during the
    transfer.
    """

    repetitions: Optional[int] = None
    volume: Optional[float] = None
    rate: Optional[float] = None


MixOpts.repetitions.__doc__ = ":py:class:`int`"
MixOpts.volume.__doc__ = ":py:class:`float`"
MixOpts.rate.__doc__ = ":py:class:`float`"


class Mix(NamedTuple):
    """
    Options to control mix behavior before aspirate and after dispense.
    """

    mix_before: MixOpts = MixOpts()
    mix_after: MixOpts = MixOpts()


Mix.mix_before.__doc__ = """
    Options applied to mix before aspirate.
    See :py:class:`.Mix.MixOpts`.
    """

Mix.mix_after.__doc__ = """
    Options applied to mix after dispense. See :py:class:`.Mix.MixOpts`.
    """


class BlowOutOpts(NamedTuple):
    """
    Location where to blow out instead of the trash.

    This location will be passed to :py:meth:`InstrumentContext.blow_out`
    when called during the transfer
    """

    location: Optional[Union[types.Location, Well]] = None


BlowOutOpts.location.__doc__ = ":py:class:`types.Location`"


class TouchTipOpts(NamedTuple):
    """
    Options to customize touch tip.

    These options will be passed to
    :py:meth:`InstrumentContext.touch_tip` when called during the
    transfer.
    """

    radius: Optional[float] = None
    v_offset: Optional[float] = None
    speed: Optional[float] = None


TouchTipOpts.radius.__doc__ = ":py:class:`float`"
TouchTipOpts.v_offset.__doc__ = ":py:class:`float`"
TouchTipOpts.speed.__doc__ = ":py:class:`float`"


class AspirateOpts(NamedTuple):
    """
    Option to customize aspirate rate.

    This option will be passed to :py:meth:`InstrumentContext.aspirate`
    when called during the transfer.
    """

    rate: Optional[float] = 1.0


AspirateOpts.rate.__doc__ = ":py:class:`float`"


class DispenseOpts(NamedTuple):
    """
    Option to customize dispense rate.

    This option will be passed to :py:meth:`InstrumentContext.dispense`
    when called during the transfer.
    """

    rate: Optional[float] = 1.0


DispenseOpts.rate.__doc__ = ":py:class:`float`"


class TransferOptions(NamedTuple):
    """
    All available options for a transfer, distribute or consolidate function
    """

    transfer: Transfer = Transfer()
    pick_up_tip: PickUpTipOpts = PickUpTipOpts()
    mix: Mix = Mix()
    blow_out: BlowOutOpts = BlowOutOpts()
    touch_tip: TouchTipOpts = TouchTipOpts()
    aspirate: AspirateOpts = AspirateOpts()
    dispense: DispenseOpts = DispenseOpts()


TransferOptions.transfer.__doc__ = """
    Options pertaining to behavior of the transfer.

    For instance you can control how frequently to get a new tip using
    :py:attr:`.Transfer.new_tip`. For documentation of all transfer options
    see :py:class:`.Transfer`.
    """

TransferOptions.pick_up_tip.__doc__ = """
    Options used when picking up a tip during transfer.
    See :py:class:`.PickUpTipOpts`.
    """

TransferOptions.mix.__doc__ = """
    Options to control mix behavior before aspirate and after dispense.
    See :py:class:`.Mix`.
    """

TransferOptions.blow_out.__doc__ = """
    Option to specify custom location for blow out. See
    :py:class:`.BlowOutOpts`.
    """

TransferOptions.touch_tip.__doc__ = """
    Options to customize touch tip. See
    :py:class:`.TouchTipOpts`.
    """

TransferOptions.aspirate.__doc__ = """
    Option to customize aspirate rate. See
    :py:class:`.AspirateOpts`.
    """

TransferOptions.dispense.__doc__ = """
    Option to customize dispense rate. See
    :py:class:`.DispenseOpts`.
    """


class TransferPlan:
    """Calculate and carry state for an arbitrary transfer

    This class encapsulates the logic around planning an M:N transfer.

    It handles calculations based on pipette channels, tip management, and all
    the various little commands that can be involved in a transfer. It can be
    iterated to resolve methods to call to execute the plan.
    """

    def __init__(
        self,
        volume,
        sources,
        dests,
        # todo(mm, 2021-03-10):
        # Refactor to not need an InstrumentContext, so we can more
        # easily test this class's logic on its own.
        instr: "InstrumentContext",
        max_volume: float,
        api_version: APIVersion,
        mode: str,
        options: Optional[TransferOptions] = None,
    ) -> None:
        """Build the transfer plan.

        This method initializes the object and does the work of preparing the
        transfer plan. Its arguments are as those of
        :py:meth:`.InstrumentContext.transfer`.
        """
        self._instr = instr
        self._api_version = api_version
        # Convert sources & dests into proper format
        # CASES:
        # i. if using multi-channel pipette,
        # and the source or target is a row/column of Wells (i.e list of Wells)
        # then avoid iterating through its Wells.
        # ii. if using single channel pipettes, flatten a multi-dimensional
        # list of Wells into a 1 dimensional list of Wells
        if self._instr.hw_pipette["channels"] > 1:
            sources, dests = self._multichannel_transfer(sources, dests)
        else:
            if isinstance(sources, List) and isinstance(sources[0], List):
                # Source is a List[List[Well]]
                sources = [well for well_list in sources for well in well_list]
            elif isinstance(sources, Well) or isinstance(sources, types.Location):
                sources = [sources]
            if isinstance(dests, List) and isinstance(dests[0], List):
                # Dest is a List[List[Well]]
                dests = [well for well_list in dests for well in well_list]
            elif isinstance(dests, Well) or isinstance(dests, types.Location):
                dests = [dests]

        total_xfers = max(len(sources), len(dests))

        self._volumes = self._create_volume_list(volume, total_xfers)
        self._sources = sources
        self._dests = dests
        self._options = options or TransferOptions()
        self._strategy = self._options.transfer
        self._tip_opts = self._options.pick_up_tip
        self._blow_opts = self._options.blow_out
        self._touch_tip_opts = self._options.touch_tip
        self._mix_before_opts = self._options.mix.mix_before
        self._mix_after_opts = self._options.mix.mix_after
        self._max_volume = max_volume

        self._mode = TransferMode[mode.upper()]

    def __iter__(self):
        if self._strategy.new_tip == types.TransferTipPolicy.ONCE:
            yield self._format_dict("pick_up_tip", kwargs=self._tip_opts)
        yield from {
            TransferMode.CONSOLIDATE: self._plan_consolidate,
            TransferMode.DISTRIBUTE: self._plan_distribute,
            TransferMode.TRANSFER: self._plan_transfer,
        }[self._mode]()
        if self._strategy.new_tip == types.TransferTipPolicy.ONCE:
            if self._strategy.drop_tip_strategy == DropTipStrategy.RETURN:
                yield self._format_dict("return_tip")
            else:
                yield self._format_dict("drop_tip")

    def _plan_transfer(self):
        """
        * **Source/ Dest:** Multiple sources to multiple destinations.
                            Src & dest should be equal length

        * **Volume:** Single volume or List of volumes is acceptable. This list
                      should be same length as sources/destinations

        * **Behavior with transfer options:**

            - New_tip: can be either NEVER or ONCE or ALWAYS
            - Air_gap: if specified, will be performed after every aspirate
            - Blow_out: can be performed after each dispense (after mix, before
                        touch_tip) at the location specified. If there is
                        liquid present in the tip (as in the case of nonzero
                        disposal volume), blow_out will be performed at either
                        user-defined location or (default) trash.
                        If no liquid is supposed to be present in the tip after
                        dispense, blow_out will be performed at dispense well
                        location (if blow out strategy is DEST)
            - Touch_tip: can be performed after each aspirate and/or after
                         each dispense
            - Mix: can be performed before aspirate and/or after dispense
                   if there is no disposal volume (i.e. can be performed
                   only when the tip is supposed to be empty)

            Considering all options, the sequence of actions is:
            *New Tip -> Mix -> Aspirate (with disposal volume) -> Air gap ->
            -> Touch tip -> Dispense air gap -> Dispense -> Mix if empty ->
            -> Blow out -> Touch tip -> Drop tip*
        """
        # reform source target lists
        sources, dests = self._extend_source_target_lists(self._sources, self._dests)
        plan_iter = self._expand_for_volume_constraints(
            self._volumes,
            zip(sources, dests),
            self._instr.max_volume
            - self._strategy.disposal_volume
            - self._strategy.air_gap,
        )
        for step_vol, (src, dest) in plan_iter:
            if self._strategy.new_tip == types.TransferTipPolicy.ALWAYS:
                yield self._format_dict("pick_up_tip", kwargs=self._tip_opts)
            max_vol = (
                self._max_volume
                - self._strategy.disposal_volume
                - self._strategy.air_gap
            )
            xferred_vol = 0.0
            while xferred_vol < step_vol:
                # TODO: account for unequal length sources, dests
                # TODO: ensure last transfer is > min_vol
                vol = min(max_vol, step_vol - xferred_vol)
                yield from self._aspirate_actions(vol, src)
                yield from self._dispense_actions(vol=vol, dest=dest, src=src)
                xferred_vol += vol
            yield from self._new_tip_action()

    @staticmethod
    def _extend_source_target_lists(
        sources: List[Union[Well, types.Location]],
        targets: List[Union[Well, types.Location]],
    ):
        """Extend source or target list to match the length of the other"""
        if len(sources) < len(targets):
            if len(targets) % len(sources) != 0:
                raise ValueError("Source and destination lists must be divisible")
            sources = [
                source
                for source in sources
                for i in range(int(len(targets) / len(sources)))
            ]
        elif len(sources) > len(targets):
            if len(sources) % len(targets) != 0:
                raise ValueError("Source and destination lists must be divisible")
            targets = [
                target
                for target in targets
                for i in range(int(len(sources) / len(targets)))
            ]
        return sources, targets

    def _plan_distribute(self):
        """
        * **Source/ Dest:** One source to many destinations
        * **Volume:** Single volume or List of volumes is acceptable. This list
                      should be same length as destinations
        * **Behavior with transfer options:**

            - New_tip: can be either NEVER or ONCE
                       (ALWAYS will fallback to ONCE)
            - Air_gap: if specified, will be performed after every aspirate and
                       also in-between dispenses (to keep air gap while moving
                       between wells)
            - Blow_out: can be performed at the end of distribute (after mix,
                        before touch_tip) at the location specified. If there
                        is liquid present in the tip, blow_out will be
                        performed at either user-defined location or (default)
                        trash. If no liquid is supposed to be present in the
                        tip at the end of distribute, blow_out will be
                        performed at the last well the liquid was dispensed to
                        (if strategy is DEST)
            - Touch_tip: can be performed after each aspirate and/or after
                         every dispense
            - Mix: can be performed before aspirate and/or after the last
                   dispense if there is no disposal volume (i.e. can be
                   performed only when the tip is supposed to be empty)

            Considering all options, the sequence of actions is:

            1. Going from source to dest1:
               *New Tip -> Mix -> Aspirate (with disposal volume) -> Air gap ->
               -> Touch tip -> Dispense air gap -> Dispense -> Mix if empty ->
               -> Blow out -> Touch tip -> Drop tip*
            2. Going from destn to destn+1:
               *.. Dispense air gap -> Dispense -> Touch tip -> Air gap ->
               .. Dispense air gap -> ...*

        """
        # TODO: decide whether default disposal vol for distribute should be
        # pipette min_vol or should we leave it to being 0 by default and
        # recommend users to specify a disposal vol when using distribute.
        # First method keeps distribute consistent with current behavior while
        # the other maintains consistency in default behaviors of all functions
        plan_iter = self._expand_for_volume_constraints(
            self._volumes,
            self._dests,
            # todo(mm, 2021-03-09): Is it right for this to be
            # _instr_.max_volume? Does/should this take the tip maximum volume
            # into account?
            self._instr.max_volume
            - self._strategy.disposal_volume
            - self._strategy.air_gap,
        )

        done = False
        current_xfer = next(plan_iter)
        if self._strategy.new_tip == types.TransferTipPolicy.ALWAYS:
            yield self._format_dict("pick_up_tip", kwargs=self._tip_opts)
        while not done:
            asp_grouped: List[Tuple[float, Well]] = []
            try:
                while (
                    sum(a[0] for a in asp_grouped)
                    + self._strategy.disposal_volume
                    + self._strategy.air_gap
                    + current_xfer[0]
                ) <= self._max_volume:
                    append_xfer = self._check_volume_not_zero(
                        self._api_version, current_xfer[0]
                    )
                    if append_xfer:
                        asp_grouped.append(current_xfer)
                    current_xfer = next(plan_iter)
            except StopIteration:
                done = True
            if not asp_grouped:
                break

            yield from self._aspirate_actions(
                sum(a[0] for a in asp_grouped) + self._strategy.disposal_volume,
                self._sources[0],
            )
            for step in asp_grouped:
                yield from self._dispense_actions(
                    vol=step[0],
                    src=self._sources[0],
                    dest=step[1],
                    is_disp_next=step is not asp_grouped[-1],
                )
        yield from self._new_tip_action()

    Target = TypeVar("Target")

    @staticmethod
    def _expand_for_volume_constraints(
        volumes: Iterator[float], targets: Iterator[Target], max_volume: float
    ) -> Generator[Tuple[float, "Target"], None, None]:
        """Split a sequence of proposed transfers if necessary to keep each
        transfer under the given max volume.
        """
        for volume, target in zip(volumes, targets):
            while volume > max_volume * 2:
                yield max_volume, target
                volume -= max_volume

            if volume > max_volume:
                volume /= 2
                yield volume, target
            yield volume, target

    def _plan_consolidate(self):
        """
        * **Source/ Dest:** Many sources to one destination
        * **Volume:** Single volume or List of volumes is acceptable. This list
                      should be same length as sources
        * **Behavior with transfer options:**

            - New_tip: can be either NEVER or ONCE
                      (ALWAYS will fallback to ONCE)
            - Air_gap: if specified, will be performed after every aspirate
                       so that the aspirated liquids do not mix inside the tip.
                       The air gap will be dispensed while dispensing the
                       liquid into the destination well.
            - Blow_out: can be performed after a dispense (after mix,
                        before touch_tip) at the location specified. If there
                        is liquid present in the tip (which shouldn't happen
                        since consolidate doesn't take a disposal vol, yet),
                        blow_out will be performed at either user-defined
                        location or (default) trash.
                        If no liquid is supposed to be present in the tip after
                        dispense, blow_out will be performed at dispense well
                        loc (if blow out strategy is DEST)
            - Touch_tip: can be performed after each aspirate and/or after
                         dispense
            - Mix: can be performed before the first aspirate and/or after
                   dispense if there is no disposal volume (i.e. can be
                   performed only when the tip is supposed to be empty)

            Considering all options, the sequence of actions is:
            1. Going from source to dest1:
               *New Tip -> Mix -> Aspirate (with disposal volume?) -> Air gap
               -> Touch tip -> Dispense air gap -> Dispense -> Mix if empty ->
               -> Blow out -> Touch tip -> Drop tip*
            2. Going from source(n) to source(n+1):
               *.. Aspirate -> Air gap -> Touch tip ->..
               .. Aspirate -> .....*
        """
        plan_iter = self._expand_for_volume_constraints(
            # todo(mm, 2021-03-09): Is it right to use _instr.max_volume here?
            # Why don't we account for tip max volume, disposal volume, or air
            # gap?
            self._volumes,
            self._sources,
            self._instr.max_volume,
        )
        current_xfer = next(plan_iter)
        if self._strategy.new_tip == types.TransferTipPolicy.ALWAYS:
            yield self._format_dict("pick_up_tip", kwargs=self._tip_opts)
        done = False
        while not done:
            asp_grouped: List[Tuple[float, Well]] = []
            try:
                while (
                    sum([a[0] for a in asp_grouped])
                    + self._strategy.disposal_volume
                    + self._strategy.air_gap * len(asp_grouped)
                    + current_xfer[0]
                ) <= self._max_volume:
                    append_xfer = self._check_volume_not_zero(
                        self._api_version, current_xfer[0]
                    )
                    if append_xfer:
                        asp_grouped.append(current_xfer)
                    current_xfer = next(plan_iter)
            except StopIteration:
                done = True
            if not asp_grouped:
                break
            # Q: What accounts as disposal volume in a consolidate action?
            # yield self._format_dict('aspirate',
            #                         self._strategy.disposal_volume, loc)
            for step in asp_grouped:
                yield from self._aspirate_actions(step[0], step[1])
            yield from self._dispense_actions(
                vol=sum([a[0] + self._strategy.air_gap for a in asp_grouped])
                - self._strategy.air_gap,
                src=None,
                dest=self._dests[0],
            )
        yield from self._new_tip_action()

    def _aspirate_actions(self, vol, loc):
        yield from self._before_aspirate(loc)
        yield self._format_dict("aspirate", [vol, loc, self._options.aspirate.rate])
        yield from self._after_aspirate()

    def _dispense_actions(self, vol, dest, src=None, is_disp_next=False):
        if self._strategy.air_gap:
            vol += self._strategy.air_gap
        yield self._format_dict("dispense", [vol, dest, self._options.dispense.rate])
        yield from self._after_dispense(dest=dest, src=src, is_disp_next=is_disp_next)

    def _before_aspirate(self, loc):
        if (
            self._strategy.mix_strategy == MixStrategy.BEFORE
            or self._strategy.mix_strategy == MixStrategy.BOTH
        ):
            if self._instr.current_volume == 0:
                mix_before_opts = self._mix_before_opts._asdict()
                mix_before_opts["location"] = loc
                yield self._format_dict("mix", kwargs=mix_before_opts)

    def _after_aspirate(self):
        if self._strategy.air_gap:
            yield self._format_dict("air_gap", [self._strategy.air_gap])
        if self._strategy.touch_tip_strategy == TouchTipStrategy.ALWAYS:
            yield self._format_dict("touch_tip", kwargs=self._touch_tip_opts)

    def _after_dispense(self, dest, src, is_disp_next=False):  # noqa: C901
        # This sequence of actions is subject to change
        if not is_disp_next:
            # If the next command is an aspirate, we are switching
            # between aspirate and dispense.
            if self._instr.current_volume == 0:
                # If we're empty, then this is when after mixes come into play
                if (
                    self._strategy.mix_strategy == MixStrategy.AFTER
                    or self._strategy.mix_strategy == MixStrategy.BOTH
                ):
                    mix_after_opts = self._mix_after_opts._asdict()
                    mix_after_opts["location"] = dest
                    yield self._format_dict("mix", kwargs=mix_after_opts)
            if self._strategy.touch_tip_strategy == TouchTipStrategy.ALWAYS:
                yield self._format_dict("touch_tip", kwargs=self._touch_tip_opts)

            if self._strategy.blow_out_strategy == BlowOutStrategy.SOURCE:
                yield self._format_dict("blow_out", [src])
            elif self._strategy.blow_out_strategy == BlowOutStrategy.DEST:
                yield self._format_dict("blow_out", [dest])
            elif self._strategy.blow_out_strategy == BlowOutStrategy.CUSTOM_LOCATION:
                yield self._format_dict("blow_out", kwargs=self._blow_opts)
            elif (
                self._strategy.blow_out_strategy == BlowOutStrategy.TRASH
                or self._strategy.disposal_volume
            ):
                yield self._format_dict(
                    "blow_out", [self._instr.trash_container.wells()[0]]
                )
        else:
            # Used by distribute
            if self._strategy.air_gap:
                yield self._format_dict("air_gap", [self._strategy.air_gap])
            if self._strategy.touch_tip_strategy == TouchTipStrategy.ALWAYS:
                yield self._format_dict("touch_tip", kwargs=self._touch_tip_opts)

    def _new_tip_action(self):
        if self._strategy.new_tip == types.TransferTipPolicy.ALWAYS:
            if self._strategy.drop_tip_strategy == DropTipStrategy.RETURN:
                yield self._format_dict("return_tip")
            else:
                yield self._format_dict("drop_tip")

    def _format_dict(
        self,
        method: str,
        args: List = None,
        kwargs: Union["Dictable", Dict[str, Any]] = None,
    ):
        if kwargs:
            if isinstance(kwargs, Dict):
                params = {key: val for key, val in kwargs.items() if val}
            else:
                params = {key: val for key, val in kwargs._asdict().items() if val}
        else:
            params = {}
        if not args:
            args = []
        return {"method": method, "args": args, "kwargs": params}

    def _create_volume_list(self, volume, total_xfers):
        if isinstance(volume, (float, int)):
            return [volume] * total_xfers
        elif isinstance(volume, tuple):
            return self._create_volume_gradient(
                volume[0], volume[-1], total_xfers, self._strategy.gradient_function
            )
        else:
            if not isinstance(volume, List):
                raise TypeError(
                    "Volume expected as a number or List or"
                    " tuple but got {}".format(volume)
                )
            elif not len(volume) == total_xfers:
                raise RuntimeError(
                    "List of volumes should be equal to number " "of transfers"
                )
            return volume

    def _create_volume_gradient(self, min_v, max_v, total, gradient=None):

        diff_vol = max_v - min_v

        def _map_volume(i):
            nonlocal diff_vol, total
            rel_x = i / (total - 1)
            rel_y = gradient(rel_x) if gradient else rel_x
            return (rel_y * diff_vol) + min_v

        return [_map_volume(i) for i in range(total)]

    def _check_valid_well_list(self, well_list, id, old_well_list):
        if self._api_version >= APIVersion(2, 2) and len(well_list) < 1:
            raise RuntimeError(
                f"Invalid {id} for multichannel transfer: {old_well_list}"
            )

    @staticmethod
    def _check_volume_not_zero(api_version: APIVersion, volume: float) -> bool:
        # We should only be adding volumes to transfer plans if it is
        # greater than zero to prevent extraneous robot movements.
        if api_version < APIVersion(2, 8):
            return True
        elif volume > 0:
            return True
        return False

    def _multichannel_transfer(self, s, d):
        # TODO: add a check for container being multi-channel compatible?
        # Helper function for multi-channel use-case
        assert (
            isinstance(s, Well)
            or isinstance(s, types.Location)
            or (isinstance(s, List) and isinstance(s[0], Well))
            or (isinstance(s, List) and isinstance(s[0], List))
            or (isinstance(s, List) and isinstance(s[0], types.Location))
        ), "Source should be a Well or List[Well] but is {}".format(s)
        assert (
            isinstance(d, Well)
            or isinstance(d, types.Location)
            or (isinstance(d, List) and isinstance(d[0], Well))
            or (isinstance(d, List) and isinstance(d[0], List))
            or (isinstance(d, List) and isinstance(d[0], types.Location))
        ), "Target should be a Well or List[Well] but is {}".format(d)

        # TODO: Account for cases where a src/dest list has a non-first-row
        # well (eg, 'B1') and would expect the robot/pipette to
        # understand that it is referring to the whole first column
        if isinstance(s, List) and isinstance(s[0], List):
            # s is a List[List]]; flatten to 1D list
            s = [well for list_elem in s for well in list_elem]
        elif isinstance(s, Well) or isinstance(s, types.Location):
            s = [s]
        new_src = []
        for well in s:
            if self._is_valid_row(well):
                new_src.append(well)
        self._check_valid_well_list(new_src, "source", s)

        if isinstance(d, List) and isinstance(d[0], List):
            # s is a List[List]]; flatten to 1D list
            d = [well for list_elem in d for well in list_elem]
        elif isinstance(d, Well) or isinstance(d, types.Location):
            d = [d]
        new_dst = []
        for well in d:
            if self._is_valid_row(well):
                new_dst.append(well)
        self._check_valid_well_list(new_dst, "target", d)
        return new_src, new_dst

    def _is_valid_row(self, well: Union[Well, types.Location]):
        if isinstance(well, types.Location):
            test_well = well.labware.as_well()
        else:
            test_well = well

        if self._api_version < APIVersion(2, 2):
            return test_well in test_well.parent.rows()[0]
        else:
            # Allow the first 2 rows to be accessible to 384-well plates;
            # otherwise, only the first row is accessible
            if test_well.parent.parameters["format"] == "384Standard":
                valid_wells = [
                    well for row in test_well.parent.rows()[:2] for well in row
                ]
                return test_well in valid_wells
            else:
                return test_well in test_well.parent.rows()[0]



"""Common pipetting command base models."""
from pydantic import BaseModel, Field

from ..types import WellLocation


class BasePipettingData(BaseModel):
    """Base class for data payloads of commands that interact with wells."""

    pipetteId: str = Field(
        ...,
        description="Identifier of pipette to use for liquid handling.",
    )
    labwareId: str = Field(
        ...,
        description="Identifier of labware to use.",
    )
    wellName: str = Field(
        ...,
        description="Name of well to use in labware.",
    )


class BaseLiquidHandlingData(BasePipettingData):
    """Base class for data payloads of commands that handle liquid."""

    volume: float = Field(
        ...,
        description="Amount of liquid in uL. Must be greater than 0 and less "
        "than a pipette-specific maximum volume.",
        gt=0,
    )
    wellLocation: WellLocation = Field(
        ...,
        description="Relative well location at which to perform the operation",
    )

    # todo(mm, 2021-03-26): This class or one of its subclasses should have a
    # field for liquid flow rate in microliters per second.
    # See Opentrons/opentrons#4837 for terminology concerns.


class BaseLiquidHandlingResult(BaseModel):
    """Base properties of a liquid handling result."""

    volume: float = Field(
        ...,
        description="Amount of liquid in uL handled in the operation.",
        gt=0,
    )



"""Factory for Python Protocol API instances."""
import asyncio
from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocol_engine.clients import SyncClient, ChildThreadTransport
from opentrons.protocol_api_experimental import ProtocolContext


class PythonContextCreator:
    """A factory to build Python ProtocolContext instances."""

    @staticmethod
    def create(protocol_engine: ProtocolEngine) -> ProtocolContext:
        """Create a fresh ProtocolContext wired to a ProtocolEngine."""
        loop = asyncio.get_running_loop()
        transport = ChildThreadTransport(engine=protocol_engine, loop=loop)
        client = SyncClient(transport=transport)

        return ProtocolContext(engine_client=client)



import typing
from . import file_operators as io, types as local_types


MAX_VERSION = 1


def check_index_version(index_path: local_types.StrPath):
    try:
        index_file = io.read_cal_file(str(index_path))
        version = index_file.get("version", 0)
        if version == 0:
            migrate_index_0_to_1(index_path)
    except FileNotFoundError:
        pass


def migrate_index_0_to_1(index_path: local_types.StrPath):
    """
    Previously, the index file was keyed as
    ```
    uri: {id: hash,
          slot: hash+parent,
          module: {{moduletype}: '{slot}-{moduletype}'}
    ```
    Now, the format is saved as
    ```
    labware_hash : {
        uri: uri,
        slot: hash+parent,
        module: {
            parent: {moduletype},
            fullParent: {slot}-{moduletype}}
    ```
    This function ensures any index files are migrated over to
    the correct format so users do not lose their calibrations
    """
    index_file = io.read_cal_file(str(index_path))
    updated_entries: typing.Dict = {}
    for key, data in index_file.items():
        uri = key
        full_hash = data["slot"]
        if data["module"]:
            parent, full_parent = list(data["module"].items())[0]
            module = {"parent": parent, "fullParent": full_parent}
        else:
            module = {}
        updated_entries[full_hash] = {
            "uri": f"{uri}",
            "slot": full_hash,
            "module": module,
        }
    migrated_file = {"version": 1, "data": updated_entries}
    io.save_to_file(index_path, migrated_file)



from enum import Enum
from dataclasses import dataclass
from typing import (
    Dict,
    NamedTuple,
    Callable,
    Any,
    Type,
    TypeVar,
    Tuple,
    Awaitable,
    Mapping,
    Union,
    TYPE_CHECKING,
)
from pathlib import Path

from opentrons.drivers.rpi_drivers.types import USBPort

if TYPE_CHECKING:
    from opentrons_shared_data.module.dev_types import (
        ThermocyclerModuleType,
        MagneticModuleType,
        TemperatureModuleType,
    )

ThermocyclerStep = Dict[str, float]

InterruptCallback = Callable[[str], None]

UploadFunction = Callable[[str, str, Dict[str, Any]], Awaitable[Tuple[bool, str]]]

LiveData = Mapping[str, Union[str, Mapping[str, Union[float, str, None]]]]


E = TypeVar("E", bound="_ProvideLookup")


class _ProvideLookup(Enum):
    @classmethod
    def from_str(cls: Type[E], typename: str) -> "E":
        for m in cls.__members__.values():
            if m.value == typename:
                return m
        raise AttributeError(f"No such type {typename}")


class ModuleType(_ProvideLookup):
    THERMOCYCLER: "ThermocyclerModuleType" = "thermocyclerModuleType"
    TEMPERATURE: "TemperatureModuleType" = "temperatureModuleType"
    MAGNETIC: "MagneticModuleType" = "magneticModuleType"


class MagneticModuleModel(_ProvideLookup):
    MAGNETIC_V1: str = "magneticModuleV1"
    MAGNETIC_V2: str = "magneticModuleV2"


class TemperatureModuleModel(_ProvideLookup):
    TEMPERATURE_V1: str = "temperatureModuleV1"
    TEMPERATURE_V2: str = "temperatureModuleV2"


class ThermocyclerModuleModel(_ProvideLookup):
    @classmethod
    def from_str(
        cls: Type["ThermocyclerModuleModel"], typename: str
    ) -> "ThermocyclerModuleModel":
        return super().from_str(typename)

    THERMOCYCLER_V1: str = "thermocyclerModuleV1"


@dataclass
class ModuleAtPort:
    port: str
    name: str
    usb_port: USBPort = USBPort(name="", sub_names=[])


class BundledFirmware(NamedTuple):
    """Represents a versioned firmware file, generally bundled into the fs"""

    version: str
    path: Path

    def __repr__(self) -> str:
        return f"<BundledFirmware {self.version}, path={self.path}>"


class UpdateError(RuntimeError):
    pass


class ModuleInfo(NamedTuple):
    model: str  # A module model such as "magneticModuleV2"
    fw_version: str  # The version of the firmware
    hw_revision: str  # the revision of the hardware
    serial: str  # the serial number


ModuleModel = Union[
    MagneticModuleModel, TemperatureModuleModel, ThermocyclerModuleModel
]


class TemperatureStatus(str, Enum):
    HOLDING = "holding at target"
    COOLING = "cooling"
    HEATING = "heating"
    IDLE = "idle"
    ERROR = "error"



""" Functions and variables for the smoothie update process """
import os
import shutil


def _ensure_programmer_executable():
    """Find the lpc21isp executable and ensure it is executable"""
    # Find the lpc21isp executable, explicitly allowing the case where it
    # is not executable (since that’s exactly what we’re trying to fix)
    updater_executable = shutil.which("lpc21isp", mode=os.F_OK)
    assert updater_executable, "could not find lpc21isp"
    # updater_executable might be None; we’re passing it here unchecked
    # because if it is None, we’re about to fail when we try to program
    # the smoothie, and we want the exception to bubble up.
    os.chmod(updater_executable, 0o777)



import logging
from collections import UserDict
from dataclasses import dataclass
from typing import Optional, List, Dict, TYPE_CHECKING

from opentrons import types
from opentrons.protocol_api.labware import load as load_lw, Labware
from opentrons.protocols.api_support.constants import STANDARD_DECK
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.geometry.deck_item import DeckItem
from opentrons.hardware_control.modules.types import ModuleType
from opentrons.protocols.geometry.module_geometry import (
    ModuleGeometry,
    ThermocyclerGeometry,
)
from opentrons_shared_data.deck import load as load_deck

if TYPE_CHECKING:
    from opentrons_shared_data.deck.dev_types import (
        SlotDefV2,
    )

MODULE_LOG = logging.getLogger(__name__)


# Amount of slots in a single deck row
ROW_LENGTH = 3
FIXED_TRASH_ID = "fixedTrash"


@dataclass
class CalibrationPosition:
    """
    A point on the deck of a robot that is used to calibrate
    aspects of the robot's movement system as defined by
    opentrons/shared-data/deck/schemas/2.json
    """

    id: str
    position: List[float]
    displayName: str


class Deck(UserDict):
    def __init__(self, load_name=STANDARD_DECK):
        super().__init__()
        row_offset = 90.5
        col_offset = 132.5
        for idx in range(1, 13):
            self.data[idx] = None
        self._positions = {
            idx + 1: types.Point((idx % 3) * col_offset, idx // 3 * row_offset, 0)
            for idx in range(12)
        }
        self._highest_z = 0.0
        self._definition = load_deck(load_name, 2)
        self._load_fixtures()

    def _load_fixtures(self):
        for f in self._definition["locations"]["fixtures"]:
            slot_name = self._check_name(f["slot"])  # type: ignore
            loaded_f = load_lw(
                f["labware"], self.position_for(slot_name)  # type: ignore
            )
            self.__setitem__(slot_name, loaded_f)

    @staticmethod
    def _assure_int(key: object) -> int:
        if isinstance(key, str):
            return int(key)
        elif isinstance(key, int):
            return key
        else:
            raise TypeError(type(key))

    def _check_name(self, key: object) -> int:
        should_raise = False
        try:
            key_int = Deck._assure_int(key)
        except Exception:
            MODULE_LOG.exception("Bad slot name: {}".format(key))
            should_raise = True
        should_raise = should_raise or key_int not in self.data
        if should_raise:
            raise ValueError("Unknown slot: {}".format(key))
        else:
            return key_int

    def __getitem__(self, key: types.DeckLocation) -> DeckItem:
        return self.data[self._check_name(key)]

    def __delitem__(self, key: types.DeckLocation) -> None:
        checked_key = self._check_name(key)
        old = self.data[checked_key]
        self.data[checked_key] = None
        if old:
            self.recalculate_high_z()

    def __setitem__(self, key: types.DeckLocation, val: DeckItem) -> None:
        slot_key_int = self._check_name(key)
        item = self.data.get(slot_key_int)

        overlapping_items = self.get_collisions_for_item(slot_key_int, val)
        if item is not None:
            if slot_key_int == 12:
                if FIXED_TRASH_ID in item.parameters.get("quirks", []):
                    pass
                else:
                    raise ValueError(f"Deck location {key} " "is for fixed trash only")
            else:
                raise ValueError(
                    f"Deck location {key} already"
                    f"  has an item: {self.data[slot_key_int]}"
                )
        elif overlapping_items:
            flattened_overlappers = [
                repr(item) for sublist in overlapping_items.values() for item in sublist
            ]
            raise ValueError(
                f"Could not load {val} as deck location {key} "
                "is obscured by "
                f'{", ".join(flattened_overlappers)}'
            )
        self.data[slot_key_int] = val
        self._highest_z = max(val.highest_z, self._highest_z)

    def __contains__(self, key: object) -> bool:
        try:
            key_int = self._check_name(key)
        except ValueError:
            return False
        return key_int in self.data

    def is_edge_move_unsafe(self, mount: types.Mount, target: "Labware") -> bool:
        """
        Check if slot next to target labware contains a module. Only relevant
        depending on the mount you are using and the column you are moving
        to inside of the labware.
        """
        slot = LabwareLike(target).first_parent()
        if not slot:
            return False
        if mount is types.Mount.RIGHT:
            other_labware = self.left_of(slot)
        else:
            other_labware = self.right_of(slot)

        return isinstance(other_labware, ModuleGeometry)

    def right_of(self, slot: str) -> Optional[DeckItem]:
        if int(slot) % ROW_LENGTH == 0:
            # We know we're at the right-most edge
            # of the given row
            return None
        else:
            idx = int(slot) + 1
            return self[str(idx)]

    def left_of(self, slot: str) -> Optional[DeckItem]:
        if int(slot) - 1 % ROW_LENGTH == 0:
            # We know we're at the left-most edge
            # of the given row
            return None
        idx = int(slot) - 1
        if idx < 1:
            return None
        return self[str(idx)]

    def position_for(self, key: types.DeckLocation) -> types.Location:
        key_int = self._check_name(key)
        return types.Location(self._positions[key_int], str(key))

    def recalculate_high_z(self):
        self._highest_z = 0.0
        for item in [lw for lw in self.data.values() if lw]:
            self._highest_z = max(item.highest_z, self._highest_z)

    def get_slot_definition(self, slot_name) -> "SlotDefV2":
        slots = self._definition["locations"]["orderedSlots"]
        slot_def = next((slot for slot in slots if slot["id"] == slot_name), None)
        if not slot_def:
            slot_ids = [slot["id"] for slot in slots]
            raise ValueError(
                f"slot {slot_name} could not be found,"
                f"valid deck slots are: {slot_ids}"
            )
        return slot_def

    def get_slot_center(self, slot_name) -> types.Point:
        defn = self.get_slot_definition(slot_name)
        return types.Point(
            defn["position"][0] + defn["boundingBox"]["xDimension"] / 2,
            defn["position"][1] + defn["boundingBox"]["yDimension"] / 2,
            defn["position"][2] + defn["boundingBox"]["zDimension"] / 2,
        )

    def resolve_module_location(
        self, module_type: ModuleType, location: Optional[types.DeckLocation]
    ) -> types.DeckLocation:
        dn_from_type = {
            ModuleType.MAGNETIC: "Magnetic Module",
            ModuleType.THERMOCYCLER: "Thermocycler",
            ModuleType.TEMPERATURE: "Temperature Module",
        }
        if isinstance(location, str) or isinstance(location, int):
            slot_def = self.get_slot_definition(str(location))
            compatible_modules = slot_def["compatibleModuleTypes"]
            if module_type.value in compatible_modules:
                return location
            else:
                raise AssertionError(
                    f"A {dn_from_type[module_type]} cannot be loaded"
                    f" into slot {location}"
                )
        else:
            valid_slots = [
                slot["id"]
                for slot in self.slots
                if module_type.value in slot["compatibleModuleTypes"]
            ]
            if len(valid_slots) == 1:
                return valid_slots[0]
            elif not valid_slots:
                raise ValueError(
                    "A {dn_from_type[module_type]} cannot be used with this " "deck"
                )
            else:
                raise AssertionError(
                    f"{dn_from_type[module_type]}s do not have default"
                    " location, you must specify a slot"
                )

    @property
    def highest_z(self) -> float:
        """Return the tallest known point on the deck."""
        return self._highest_z

    @property
    def slots(self) -> List["SlotDefV2"]:
        """Return the definition of the loaded robot deck."""
        return self._definition["locations"]["orderedSlots"]

    @property
    def calibration_positions(self) -> List[CalibrationPosition]:
        raw_positions = self._definition["locations"]["calibrationPoints"]
        return [CalibrationPosition(**pos) for pos in raw_positions]

    def get_calibration_position(self, id: str) -> CalibrationPosition:
        calibration_position = next(
            (pos for pos in self.calibration_positions if pos.id == id), None
        )
        if not calibration_position:
            pos_ids = [pos.id for pos in self.calibration_positions]
            raise ValueError(
                f"calibration position {id} "
                "could not be found, "
                f"valid calibration position ids are: {pos_ids}"
            )
        return calibration_position

    def get_fixed_trash(self) -> Optional[Labware]:
        fixtures = self._definition["locations"]["fixtures"]
        ft = next((f for f in fixtures if f["id"] == FIXED_TRASH_ID), None)
        return self.data[self._check_name(ft.get("slot"))] if ft else None

    def get_non_fixture_slots(self) -> List[types.DeckLocation]:
        fixtures = self._definition["locations"]["fixtures"]
        fixture_slots = {
            self._check_name(f.get("slot")) for f in fixtures if f.get("slot")
        }
        return [s for s in self.data.keys() if s not in fixture_slots]

    def get_collisions_for_item(
        self, slot_key: types.DeckLocation, item: DeckItem
    ) -> Dict[types.DeckLocation, List[DeckItem]]:
        """Return the loaded deck items that collide
        with the given item.
        """

        def get_item_covered_slot_keys(sk, i):
            if isinstance(i, ThermocyclerGeometry):
                return i.covered_slots
            elif i is not None:
                return set([sk])
            else:
                return set([])

        item_slot_keys = get_item_covered_slot_keys(slot_key, item)

        colliding_items: Dict[types.DeckLocation, List[DeckItem]] = {}
        for sk, i in self.data.items():
            covered_sks = get_item_covered_slot_keys(sk, i)
            if item_slot_keys.issubset(covered_sks):
                colliding_items.setdefault(sk, []).append(i)
        return colliding_items



import asyncio
import logging
from opentrons.hardware_control.emulation.connection_handler import ConnectionHandler
from opentrons.hardware_control.emulation.magdeck import MagDeckEmulator
from opentrons.hardware_control.emulation.parser import Parser
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.emulation.tempdeck import TempDeckEmulator
from opentrons.hardware_control.emulation.thermocycler import ThermocyclerEmulator
from opentrons.hardware_control.emulation.smoothie import SmoothieEmulator

logger = logging.getLogger(__name__)


SMOOTHIE_PORT = 9996
THERMOCYCLER_PORT = 9997
TEMPDECK_PORT = 9998
MAGDECK_PORT = 9999


class ServerManager:
    """
    Class to start and stop emulated smoothie and modules.
    """

    def __init__(self, settings=Settings()) -> None:
        host = settings.host
        self._mag_emulator = MagDeckEmulator(parser=Parser())
        self._temp_emulator = TempDeckEmulator(parser=Parser())
        self._therm_emulator = ThermocyclerEmulator(parser=Parser())
        self._smoothie_emulator = SmoothieEmulator(
            parser=Parser(), settings=settings.smoothie
        )

        self._mag_server = self._create_server(
            host=host,
            port=MAGDECK_PORT,
            handler=ConnectionHandler(self._mag_emulator),
        )
        self._temp_server = self._create_server(
            host=host,
            port=TEMPDECK_PORT,
            handler=ConnectionHandler(self._temp_emulator),
        )
        self._therm_server = self._create_server(
            host=host,
            port=THERMOCYCLER_PORT,
            handler=ConnectionHandler(self._therm_emulator),
        )
        self._smoothie_server = self._create_server(
            host=host,
            port=SMOOTHIE_PORT,
            handler=ConnectionHandler(self._smoothie_emulator),
        )

    async def run(self):
        await asyncio.gather(
            self._mag_server,
            self._temp_server,
            self._therm_server,
            self._smoothie_server,
        )

    @staticmethod
    async def _create_server(host: str, port: int, handler: ConnectionHandler) -> None:
        """Run a server."""
        server = await asyncio.start_server(handler, host, port)

        async with server:
            await server.serve_forever()

    def reset(self):
        self._smoothie_emulator.reset()
        self._mag_emulator.reset()
        self._temp_emulator.reset()
        self._therm_emulator.reset()

    def stop(self):
        self._smoothie_server.close()
        self._temp_server.close()
        self._therm_server.close()
        self._mag_server.close()


if __name__ == "__main__":
    logging.basicConfig(format="%(asctime)s:%(message)s", level=logging.DEBUG)
    asyncio.run(ServerManager().run())



from opentrons.config import advanced_settings as advs


def short_fixed_trash() -> bool:
    return advs.get_setting_with_env_overload("shortFixedTrash")


def calibrate_to_bottom() -> bool:
    return advs.get_setting_with_env_overload("calibrateToBottom")


def dots_deck_type() -> bool:
    return advs.get_setting_with_env_overload("deckCalibrationDots")


def disable_home_on_boot() -> bool:
    return advs.get_setting_with_env_overload("disableHomeOnBoot")


def use_old_aspiration_functions() -> bool:
    return advs.get_setting_with_env_overload("useOldAspirationFunctions")


def enable_door_safety_switch() -> bool:
    return advs.get_setting_with_env_overload("enableDoorSafetySwitch")


def enable_http_protocol_sessions() -> bool:
    return advs.get_setting_with_env_overload("enableHttpProtocolSessions")


def enable_protocol_engine() -> bool:
    """Get if the ProtocolEngine should be used to run protocol files."""

    return advs.get_setting_with_env_overload("enableProtocolEngine")


def disable_fast_protocol_upload() -> bool:
    return advs.get_setting_with_env_overload("disableFastProtocolUpload")



import logging

from opentrons.drivers.asyncio.communication import SerialConnection


log = logging.getLogger(__name__)


class RemoveUnwantedCharactersMixin:
    """Mixin that removes echoed command and other unwanted characters
    from the response."""

    def process_raw_response(self, command: str, response: str) -> str:
        """Process the raw response."""
        return self._remove_unwanted_characters(command=command, response=response)

    @staticmethod
    def _remove_unwanted_characters(command: str, response: str) -> str:
        # smoothieware can enter a weird state, where it repeats back
        # the sent command at the beginning of its response.
        # Check for this echo, and strips the command from the response
        def _is_token_command(_s: str) -> bool:
            """check if token is a command"""
            # A single letter token cannot be assumed to be a command.
            # For example: "M369 L" response is "L:2132121212".
            return len(_s) > 1

        # Split at spaces.
        tokens = (c.strip() for c in command.strip().split(" "))
        # A list of commands to remove from response. Including the entire
        # command.
        remove_from_response = [command] + [c for c in tokens if _is_token_command(c)]

        # also removing any inadvertent newline/return characters
        # this is ok because all data we need from Smoothie is returned on
        # the first line in the response
        remove_from_response += ["\r", "\n"]
        modified_response = str(response)

        for cmd in remove_from_response:
            modified_response = modified_response.replace(cmd, "")

        if modified_response != response:
            log.debug(f"Removed characters from response: {response}")
            log.debug(f"Newly formatted response: {modified_response}")

        return modified_response.strip()


class SmoothieConnection(RemoveUnwantedCharactersMixin, SerialConnection):
    pass
